from __future__ import annotations

from simple_websocket.ws import ConnectionClosed
import flask_sock, typing, traceback

from .execution import Execution, RequestData, NoExecutionFound
from ..contract import should_send, forms_contract
from ..utils import deserialize, serialize
from ..contract import forms_contract
from ..jwt_auth import UserClaims


if typing.TYPE_CHECKING:
    from ..repositories.project.project import FormStage


class FormExecution(Execution):
    type = "execution"
    _connection: flask_sock.Server

    @staticmethod
    def get_current_execution() -> typing.Optional["FormExecution"]:
        execution = Execution.get_current_execution()
        if isinstance(execution, FormExecution):
            return execution

        return None

    @staticmethod
    def broadcast(msg: forms_contract.Message):
        for e in list(Execution.executions.values()):
            if isinstance(e, FormExecution) and e.connected and e.closed == False:
                e.send(msg)

    def __init__(
        self,
        runtime_json: "FormStage",
        is_initial: bool,
        connection: flask_sock.Server,
        request: RequestData,
        execution_id=None,
    ):
        self._connection = connection
        if execution_id is not None:
            self.id = execution_id

        super().__init__(runtime_json, is_initial, request, execution_id=execution_id)

    def send(self, msg: forms_contract.Message):
        self.log(msg.type, msg.data)

        if not should_send(msg, self.is_preview):
            return

        str_data = serialize(msg.to_json(self.is_preview))
        self._connection.send(str_data)

    def recv(self) -> typing.Tuple[str, typing.Dict]:
        try:
            str_data = self._connection.receive()
        except ConnectionClosed:
            self.close()
            return "close", {}

        data = deserialize(str_data)
        self.log(data["type"], data)
        return data["type"], data

    def close(self, reason: typing.Optional[str] = ""):
        self._connection.close(message=reason)
        self.closed = True

    def stdio(self, type: str, text: str):
        self.send(forms_contract.StdioMessage(type, text))
        return super().stdio(type, text)

    @property
    def connected(self) -> bool:
        return self._connection.connected

    def end(self):
        if self.connected:
            self.close(reason=traceback.format_exc() or None)

    @property
    def query_params(self) -> typing.Dict:
        return self.context.get("query_params", {})

    @query_params.setter
    def query_params(self, value: typing.Dict) -> None:
        self.context["query_params"] = value

    def receive(self):
        while True:
            type, data = self.recv()
            if type in ["heartbeat", "browser:try-disconnect"]:
                continue

            return data

    def get_user(self, refresh: bool = False):
        # TODO: store/cache this
        self.send(forms_contract.AuthRequireInfoMessage(refresh=refresh))

        while True:
            type, data = self.recv()
            if type != "auth:saved-jwt":
                continue

            claims = UserClaims.from_jwt(data["jwt"])
            if not claims:
                self.send(forms_contract.AuthInvalidJWTMessage())
                continue

            self.send(forms_contract.AuthValidJWTMessage())
            return claims

    def execute_js(self, code: str, context: dict = {}):
        self.send(forms_contract.ExecuteJSRequestMessage(code, context))

        while True:
            type, data = self.recv()
            if type != "execute-js:response":
                continue

            return data.get("value")

    def alert(self, message: str, severity: str):
        severity = (
            severity if severity in ["info", "warn", "error", "success"] else "info"
        )
        self.send(forms_contract.AlertMessage(message, severity))

    def redirect(self, url: str, query_params: dict):
        query_params = query_params or self.query_params
        self.send(forms_contract.RedirectMessage(url, query_params))

    # forms

    def _wait_start(self):
        type = None
        data = None

        type, data = self.recv()

        if type != "start":
            return self._wait_start()

        return data["params"]

    def setup_context(self, request: RequestData) -> None:
        self.query_params = self._wait_start()
        self.send(forms_contract.ExecutionIdMessage(self.id))

    def handle_success(self) -> None:
        close_dto = forms_contract.CloseDTO(exit_status="SUCCESS")
        self.send(forms_contract.CloseMessage(close_dto))

    def _handle_ws_exception_1001(self, exception: flask_sock.ConnectionClosed) -> None:
        self.log(
            "connection-closed",
            {"message": "Client went away - probably closed the tab."},
        )

    def _handle_ws_exception_other(
        self, exception: flask_sock.ConnectionClosed
    ) -> None:
        self.log(
            "connection-closed",
            {
                "message": f"[ERROR] Connection closed with code {exception.reason}: {exception.message}\n",
                "reason": exception.reason,
            },
        )

    def _handle_ws_exception(self, exception: flask_sock.ConnectionClosed) -> None:
        if exception.reason == 1000:
            return super().handle_success()  # missing advanve steps?

        if exception.reason == 1001:
            return self._handle_ws_exception_1001(exception)

        return self._handle_ws_exception_other(exception)

    def handle_failure(self, exception: Exception) -> None:
        if isinstance(exception, flask_sock.ConnectionClosed):
            self._handle_ws_exception(exception)

        close_dto = forms_contract.CloseDTO(
            exit_status="GENERIC_EXCEPTION",
            exception=exception.__str__(),
        )
        self.send(forms_contract.CloseMessage(close_dto))

        return super().handle_failure(exception)

    def handle_finish(self):
        self.close()


def get_form_execution_throwable() -> FormExecution:
    execution = FormExecution.get_current_execution()
    if not execution:
        raise NoExecutionFound()
    return execution



import flask

from ...usage import usage
from .main import MainController


def get_editor_bp(controller: MainController):
    bp = flask.Blueprint("editor_workflows", __name__)

    @bp.get("/initial-data")
    @usage
    def _get_workflow_editor_data():
        try:
            return controller.get_workflow_editor_data()
        except Exception as e:
            return str(e), 500

    @bp.post("/move")
    @usage
    def _workflow_move():
        try:
            payload = flask.request.json
            controller.workflow_move(payload)
            return flask.Response(status=204)
        except Exception as e:
            return str(e), 500

    @bp.post("/add-nodes")
    @usage
    def _bulk_create_stages():
        try:
            payload = flask.request.json
            controller.bulk_create_stages(payload)
            return flask.Response(status=204)
        except Exception as e:
            return str(e), 500

    @bp.post("/duplicate-nodes")
    @usage
    def _bulk_duplicate_stages():
        try:
            payload = flask.request.json
            controller.bulk_duplicate_stages(payload)
            return flask.Response(status=204)
        except Exception as e:
            return str(e), 500

    @bp.post("/delete")
    @usage
    def _bulk_delete():
        try:
            payload = flask.request.json
            controller.bulk_delete(payload)
            return flask.Response(status=204)
        except Exception as e:
            return str(e), 500

    @bp.post("/add-transition")
    @usage
    def _bulk_create_transitions():
        try:
            payload = flask.request.json
            controller.bulk_create_transitions(payload)
            return flask.Response(status=204)
        except Exception as e:
            return str(e), 500

    return bp



from .LatexOutput import LatexOutput
from .CardsInput import CardsInput
from .ToggleInput import ToggleInput
from .ChecklistInput import ChecklistInput
from .TextOutput import TextOutput
from .CnpjInput import CnpjInput
from .CustomInput import CustomInput
from .MultipleChoiceInput import MultipleChoiceInput
from .PhoneInput import PhoneInput
from .NpsInput import NpsInput
from .TimeInput import TimeInput
from .DropdownInput import DropdownInput
from .RatingInput import RatingInput
from .VideoInput import VideoInput
from .PandasRowSelectionInput import PandasRowSelectionInput
from .LinkOutput import LinkOutput
from .CurrencyInput import CurrencyInput
from .RichTextInput import RichTextInput
from .TagInput import TagInput
from .EmailInput import EmailInput
from .PandasOutput import PandasOutput
from .DateInput import DateInput
from .IframeOutput import IframeOutput
from .NumberSliderInput import NumberSliderInput
from .HtmlOutput import HtmlOutput
from .PlotlyOutput import PlotlyOutput
from .TextInput import TextInput
from .ImageInput import ImageInput
from .PasswordInput import PasswordInput
from .ListInput import ListInput
from .MarkdownOutput import MarkdownOutput
from .CodeInput import CodeInput
from .ProgressOutput import ProgressOutput
from .NumberInput import NumberInput
from .CpfInput import CpfInput
from .CheckboxInput import CheckboxInput
from .FileOutput import FileOutput
from .FileInput import FileInput
from .CameraInput import CameraInput
from .TextareaInput import TextareaInput
from .ImageOutput import ImageOutput



import debugpy, json
from ...utils import get_free_port
from ...settings import Settings
from ...templates import vscode_launch_json

CONNECTED_PORT = None


def is_launch_json_configured():
    launch_json_file = Settings.root_path.joinpath(".vscode/launch.json")
    if not launch_json_file.exists():
        return False

    # TODO: Check if launch.json is configured correctly
    return True


def configure_launch_json():
    Settings.root_path.joinpath(".vscode").mkdir(exist_ok=True)
    Settings.root_path.joinpath(".vscode/launch.json").write_text(
        json.dumps(vscode_launch_json, indent=4), encoding="utf-8"
    )


def start_debugger():
    global CONNECTED_PORT
    if CONNECTED_PORT is None:
        CONNECTED_PORT = get_free_port(5678)
        debugpy.listen(CONNECTED_PORT)
    return CONNECTED_PORT


def is_client_connected():
    return debugpy.is_client_connected()



from ..widget_base import Input
from io import BufferedReader, TextIOWrapper
from typing import Optional, Union, List
from ..apis import upload_file
from ..response_types import FileResponse
from pathlib import Path


class CameraInput(Input):
    type = "camera-input"
    empty_value = None

    def __init__(self, key: str, label: str, **kwargs):
        super().__init__(key)
        self.set_props(dict(label=label, **kwargs))

    def set_props(self, props):
        self.label = props.get("label", "Label")
        self.required = props.get("required", True)
        self.hint = props.get("hint", None)
        self.full_width = props.get("full_width", False)
        self.layout = props.get("layout", "list")
        self.disabled = props.get("disabled", False)
        self.value = props.get("initial_value", self.empty_value)

    def render(self, context: dict):
        return {
            "type": self.type,
            "key": self.key,
            "label": self.label,
            "hint": self.hint,
            "value": self.serialize_value(),
            "required": self.required,
            "fullWidth": self.full_width,
            "layout": self.layout,
            "disabled": self.disabled,
            "errors": self.errors,
        }

    @staticmethod
    def __get_file_uri(
        value: Union[FileResponse, str, BufferedReader, TextIOWrapper]
    ) -> str:
        if isinstance(value, str):
            if Path(value).is_file():
                return upload_file(open(value))
            else:
                return value
        if isinstance(value, FileResponse):
            return value.url
        if isinstance(value, BufferedReader):
            return upload_file(value)
        if isinstance(value, TextIOWrapper):
            return upload_file(value)
        return ""

    def serialize_value(self):
        if self.value is None:
            return None
        return CameraInput.__get_file_uri(self.value)

    def parse_value(
        self, value: Optional[List[str]]
    ) -> Optional[Union[FileResponse, List[FileResponse]]]:
        if value is None:
            return self.empty_value
        file_response = FileResponse(value)
        return super().parse_value(file_response)



from __future__ import annotations  # Required for TYPE_CHECKING
from typing import TYPE_CHECKING, Any, ClassVar, Dict, List, Optional
from dataclasses import dataclass
import threading
import traceback
import uuid

from ..modules import import_as_new
from ..monitoring import LogMessage, log
from ..utils.environment import IS_PREVIEW
from .stoppable_thread import StoppableThread

if TYPE_CHECKING:
    from ..repositories.project.project import WorkflowStage


class NoExecutionFound(Exception):
    def __str__(self) -> str:
        return "No execution found. If you are running this locally, make sure you are running it with `abstra serve`. For more information, see https://docs.abstra.io"


@dataclass
class RequestData:
    headers: Dict[str, str]
    body: str
    method: str
    query_params: Dict[str, str]


class Execution:
    type = "execution"
    abstra_run_key = "abstra-run-id"
    executions: ClassVar[Dict[int, "Execution"]] = {}

    id: str
    is_initial: bool
    request: RequestData
    thread: StoppableThread

    @classmethod
    def get_execution_by_id(cls, execution_id: str) -> Optional["Execution"]:
        for execution in cls.executions.values():
            if execution.id == execution_id:
                return execution
        return None

    @classmethod
    def get_current_execution(cls) -> Optional["Execution"]:
        thread_id = threading.get_ident()
        return cls.executions.get(thread_id)

    def __init__(
        self,
        runtime_json: WorkflowStage,
        is_initial: bool,
        request: RequestData,
        execution_id=None,
    ):
        if execution_id:
            self.id = execution_id
        else:
            self.id = uuid.uuid4().__str__()

        self.is_initial = is_initial

        self.thread = StoppableThread(target=self._run, args=())
        self.request = request

        self.stderr: List[str] = []
        self.stdout: List[str] = []
        self.context: Dict = {}
        self.stage = runtime_json

    @property
    def is_preview(self) -> bool:
        return IS_PREVIEW

    def run_async(self):
        self.thread.start()

    def stop(self):
        self.thread.kill()

    def run_sync(self):
        self.thread.start()
        self.thread.join()

    def stdio(self, type: str, text: str):
        if type == "stderr":
            self.stderr.append(text)
        elif type == "stdout":
            self.stdout.append(text)

    def log(self, event: str, payload: Any):
        if event == "heartbeat":
            return
        log(
            LogMessage(
                event=event,
                payload=payload,
                executionId=self.id,
                runtime_type=self.stage.runner_type,
                runtime_name=self.stage.title,
            )
        )

    def handle_started(self) -> None:
        pass

    def handle_success(self) -> None:
        pass

    def handle_failure(self, e: Exception) -> None:
        traceback.print_exc()

    def setup_context(self, request: RequestData):
        raise NotImplementedError()

    def _run(self) -> None:
        self.thread_id = threading.get_ident()
        Execution.executions[self.thread_id] = self

        self.setup_context(self.request)
        self.handle_started()

        try:
            try:
                import_as_new(self.stage.file)
            except SystemExit as e:
                if e.code is not None and e.code != 0:
                    raise e
            self.handle_success()
        except Exception as e:
            self.handle_failure(e)
        finally:
            self.delete()

    def delete(self) -> None:
        del Execution.executions[self.thread_id]


def get_current_execution_throwable() -> Execution:
    execution = Execution.get_current_execution()
    if not execution:
        raise NoExecutionFound()
    return execution
